@@ -1,4 +1,6 @@
-use strict;
+#!/usr/bin/perl
+
+use v5.14;
 use warnings;
 
 package LispParser;
